import argparse

from . import util
from .util import DockerObjectNotFound

try:
    from . import Atomic
except ImportError:
    from atomic import Atomic # pylint: disable=relative-import

def cli(subparser):
    # atomic stop
    stopp = subparser.add_parser(
        "stop", help=_("execute container image stop method"),
        epilog="atomic will just stop the container if it is running, if "
        "image does not specify LABEL STOP")
    stopp.set_defaults(_class=Stop, func='stop')
    util.add_opt(stopp)
    stopp.add_argument("-n", "--name", dest="name", default=None,
                       help=_("name of container"))
    stopp.add_argument("image", help=_("container image"))
    stopp.add_argument("args", nargs=argparse.REMAINDER,
                          help=_("Additional arguments appended to the image "
                                 "stop method"))

class Stop(Atomic):
    def __init__(self):
        super(Stop, self).__init__()

    def stop(self):
        if self.syscontainers.get_checkout(self.name) is not None:
            self.syscontainers.stop_service(self.name)
            return

        self.inspect = self._inspect_container()
        if self.inspect is None:
            self.inspect = self._inspect_image()
            if self.inspect is None:
                raise DockerObjectNotFound(self.name)

        args = self._get_args("STOP")
        if args:
            cmd = self.gen_cmd(args + self.quote(self.args.args))
            cmd = self.sub_env_strings(cmd)
            self.display(cmd)
            util.check_call(cmd, env=self.cmd_env())

        # Container exists
        try:
            if self.inspect["State"]["Running"]:
                self.d.stop(self.name)
        except KeyError:
            pass



import argparse
import sys
from . import util
from .util import add_opt
from .syscontainers import OSTREE_PRESENT

try:
    from . import Atomic
except ImportError:
    from atomic import Atomic # pylint: disable=relative-import

INSTALL_ARGS = ["run",
                "-t",
                "-i",
                "--rm",
                "--privileged",
                "-v", "/:/host",
                "--net=host",
                "--ipc=host",
                "--pid=host",
                "-e", "HOST=/host",
                "-e", "NAME=${NAME}",
                "-e", "IMAGE=${IMAGE}",
                "-e", "CONFDIR=/host/etc/${NAME}",
                "-e", "LOGDIR=/host/var/log/${NAME}",
                "-e", "DATADIR=/host/var/lib/${NAME}",
                "--name", "${NAME}",
                "${IMAGE}"]

def cli(subparser):
    # atomic install
    installp = subparser.add_parser(
        "install", help=_("execute container image install method"),
        epilog="atomic install attempts to read the LABEL INSTALL field "
        "in the image, if it does not exist atomic will just pull "
        "the image on to your machine.  You could add a LABEL "
        "INSTALL command to your Dockerfile like: 'LABEL INSTALL "
        "%s'" % Install.print_install())
    installp.set_defaults(_class=Install, func='install')
    add_opt(installp)
    installp.add_argument("-n", "--name", dest="name", default=None,
                          help=_("name of container"))
    installp.add_argument(
        "--display",
        default=False,
        action="store_true",
        help=_("preview the command that %s would execute") % sys.argv[0])
    installp.add_argument("image", help=_("container image"))
    if OSTREE_PRESENT:
        bwrap_oci_available = util.bwrap_oci_available()
        runc_available = util.runc_available()
        if bwrap_oci_available or runc_available:
            system_xor_user = installp.add_mutually_exclusive_group()
            if bwrap_oci_available:
                system_xor_user.add_argument("--user", dest="user", action="store_true", default=False,
                                             help=_("Flag to specify if user is non-root privileged."))
            if runc_available:
                system_xor_user.add_argument("--system", dest="system",
                                             action='store_true', default=False,
                                             help=_('install a system container'))
        installp.add_argument("--rootfs", dest="remote",
                              help=_("choose an existing exploded container/image to use "
                                     "its rootfs as a remote, read-only rootfs for the "
                                     "container to be installed"))
        installp.add_argument("--set", dest="setvalues",
                              action='append',
                              help=_("Specify a variable in the VARIABLE=VALUE "
                                     "form for a system container"))
    installp.add_argument("args", nargs=argparse.REMAINDER,
                          help=_("Additional arguments appended to the image "
                                 "install method"))

class Install(Atomic):
    def __init__(self):
        super(Install, self).__init__()

    def install(self):
        if self._container_exists(self.name):
            raise ValueError("A container '%s' is already present" % self.name)

        if self.user:
            if not util.is_user_mode():
                raise ValueError("--user does not work for privileged user")
            return self.syscontainers.install_user_container(self.image, self.name)
        elif self.system:
            return self.syscontainers.install(self.image, self.name)
        elif OSTREE_PRESENT and self.args.setvalues:
            raise ValueError("--set is valid only when used with --system or --user")

        self._check_if_image_present()
        args = self._get_args("INSTALL")
        if not args:
            return

        cmd = self.sub_env_strings(self.gen_cmd(args + self.quote(self.args.args)))

        self.display(cmd)

        if not self.args.display:
            return util.check_call(cmd)

    def _check_if_image_present(self):
        self.inspect = self._inspect_image()
        if not self.inspect:
            if self.args.display:
                self.display("Need to pull %s" % self.image)
                return
            tag = util.Decompose(self.image).tag
            # skopeo requires the use of tags or it will fail
            # if not tag is found, use 'latest'
            if not tag:
                self.image += ":{}".format("latest")
            self.update()
            self.inspect = self._inspect_image()

    @staticmethod
    def print_install():
        return "%s %s %s" % (util.default_docker(), " ".join(INSTALL_ARGS), "/usr/bin/INSTALLCMD")




import unittest

from Atomic.syscontainers import SystemContainers

class TestAtomicPull(unittest.TestCase):
    class Args():
        def __init__(self):
            self.image = "docker:fedora"
            self.user = False

    def test_pull_as_privileged_user(self):
        args = self.Args()
        testobj = SystemContainers()
        testobj.set_args(args)
        testobj.pull_image()

    def test_pull_as_nonprivileged_user(self):
        args = self.Args()
        args.user = True
        testobj = SystemContainers()
        testobj.set_args(args)
        testobj.pull_image()

if __name__ == '__main__':
    unittest.main()



import sys
import os
from . import util

try:
    from . import Atomic
except ImportError:
    from atomic import Atomic # pylint: disable=relative-import

SPC_ARGS = ["run",
            "-t",
            "-i",
            "--privileged",
            "-v", "/:/host",
            "-v", "/run:/run",
            "-v", "/etc/localtime:/etc/localtime",
            "-v", "/sys/fs/selinux:/sys/fs/selinux:ro",
            "--net=host",
            "--ipc=host",
            "--pid=host",
            "-e", "HOST=/host",
            "-e", "NAME=${NAME}",
            "-e", "IMAGE=${IMAGE}",
            "--name", "${NAME}",
            "${IMAGE}"]

RUN_ARGS = ["run",
            "-t",
            "-i",
            "--name", "${NAME}",
            "${IMAGE}"]

def cli(subparser):
    # atomic run
    runp = subparser.add_parser(
        "run", help=_("execute container image run method"),
        epilog="atomic run attempts to start an existing container if a container "
        "name is specified, or execute container image run method if an image "
        "name is specified.  Defaults to the following command, if image does "
        "not specify LABEL run:\n'%s'" % Run.print_run())
    runp.set_defaults(_class=Run, func='run')
    run_group = runp.add_mutually_exclusive_group()
    util.add_opt(runp)
    runp.add_argument("-n", "--name", dest="name", default=None,
                      help=_("name of container"))
    runp.add_argument("--spc", default=False, action="store_true",
                      help=_("use super privileged container mode: '%s'" %
                             Run.print_spc()))
    runp.add_argument("image", help=_("container image"))
    runp.add_argument("command", nargs="*",
                      help=_("optional command to execute within the container. "
                             "If container is not running, command is appended "
                             "to the image run method"))

    run_group.add_argument("--quiet", "-q", action="store_true",
                      help=_("Be less verbose."))

    run_group.add_argument(
        "--display",
        default=False,
        action="store_true",
        help=_("preview the command that %s would execute") % sys.argv[0])


class Run(Atomic):
    def __init__(self):
        super(Run, self).__init__()

    def run(self):
        if self.syscontainers.get_checkout(self.name) is not None:
            self.syscontainers.start_service(self.name)
            return

        self.inspect = self._inspect_container()
        if self.inspect:
            self._check_latest()
            # Container exists
            if self.inspect["State"]["Running"]:
                return self._running()
            elif not self.args.display:
                return self._start()

        # Container does not exist
        self.inspect = self._inspect_image()
        if not self.inspect:
            if self.args.display:
                return self.display("Need to pull %s" % self.image)

            self.update()
            self.inspect = self._inspect_image()

        if self.spc:
            if self.command:
                args = [self.docker_binary()] + SPC_ARGS + self.command
            else:
                args = [self.docker_binary()] + SPC_ARGS + self._get_cmd()
        else:
            args = self._get_args("RUN")
            if args:
                args += self.command
                opts_file = self._get_args("RUN_OPTS_FILE")
                if opts_file:
                    opts_file = self.sub_env_strings("".join(opts_file))
                    if opts_file.startswith("/"):
                        if os.path.isfile(opts_file):
                            try:
                                self.run_opts = open(opts_file, "r").read()
                            except IOError:
                                raise ValueError("Failed to read RUN_OPTS_FILE %s" % opts_file)
                    else:
                        raise ValueError("Will not read RUN_OPTS_FILE %s: not absolute path" % opts_file)
            else:
                if self.command:
                    args = [self.docker_binary()] + RUN_ARGS + self.command
                else:
                    args = [self.docker_binary()] + RUN_ARGS + self._get_cmd()

        if len(args) > 0 and args[0] == "docker":
            args[0] = self.docker_binary()

        cmd = self.gen_cmd(args)
        cmd = self.sub_env_strings(cmd)
        self.display(cmd)
        if self.args.display:
            return

        if not self.args.quiet:
            self.check_args(cmd)
        util.check_call(cmd, env=self.cmd_env())

    @staticmethod
    def check_args(cmd):
        found_sec_arg = False
        security_args = {
            '--privileged':
                'This container runs without separation and should be '
                'considered the same as root on your system.',
            '--cap-add':
                'Adding capabilities to your container could allow processes '
                'from the container to break out onto your host system.',
            '--security-opt label:disable':
                'Disabling label separation turns off tools like SELinux and '
                'could allow processes from the container to break out onto '
                'your host system.',
            '--net=host':
                'Processes in this container can listen to ports (and '
                'possibly rawip traffic) on the host\'s network.',
            '--pid=host':
                'Processes in this container can see and interact with all '
                'processes on the host and disables SELinux within the '
                'container.',
            '--ipc=host':
                'Processes in this container can see and possibly interact '
                'with all semaphores and shared memory segments on the host '
                'as well as disables SELinux within the container.'
        }

        for sec_arg in security_args:
            if sec_arg in cmd:
                if not found_sec_arg:
                    util.write_out("\nThis container uses privileged "
                                  "security switches:")
                util.write_out("\n\033[1mINFO: {}\033[0m "
                              "\n{}{}".format(sec_arg, " " * 6,
                                              security_args[sec_arg]))
                found_sec_arg = True
        if found_sec_arg:
            util.write_out("\nFor more information on these switches and their "
                          "security implications, consult the manpage for "
                          "'docker run'.\n")

    @staticmethod
    def print_run():
        return "%s %s" % (util.default_docker(), " ".join(RUN_ARGS))

    @staticmethod
    def print_spc():
        return "%s %s" % (util.default_docker(), " ".join(SPC_ARGS))



try:
    from . import Atomic
except ImportError:
    from atomic import Atomic # pylint: disable=relative-import

from .syscontainers import OSTREE_PRESENT

def cli(subparser):
    # atomic update
    updatep = subparser.add_parser(
        "update", help=_("pull latest container image from repository"),
        epilog="atomic update downloads the latest container image. If a "
        "previously created  container based on this image exists, "
        "the container will continue to use the old image.  Use "
        "--force to remove the outdated container.")
    updatep.set_defaults(_class=Update, func='update')
    updatep.add_argument("-f", "--force", default=False, dest="force",
                         action="store_true",
                         help=_("remove all containers based on this image"))
    if OSTREE_PRESENT:
        updatep.add_argument("--set", dest="setvalues",
                             action='append',
                             help=_("Specify a variable in the VARIABLE=VALUE "
                                    "form for a system container"))
    updatep.add_argument("--container", dest="container",
                         action='store_true', default=False,
                         help=_('update an installed container'))
    updatep.add_argument("image", help=_("container image"))

class Update(Atomic):
    def __init__(self):
        super(Update, self).__init__()

    def update(self):
        super(Update, self).update()



from . import Atomic
from . import util
from docker.errors import NotFound
from docker.errors import APIError
import sys

class Delete(Atomic):
    def __init__(self):
        super(Delete, self).__init__()

    def delete_image(self):
        """
        Mark given image(s) for deletion from registry
        :return: 0 if all images marked for deletion, otherwise 2 on any failure
        """

        if not self.args.assumeyes:
            confirm = util.input("Do you wish to delete {}? (y/N) ".format(self.args.delete_targets))
            confirm = confirm.strip().lower()
            if not confirm in ['y', 'yes']:
                util.write_err("User aborted delete operation for {}".format(self.args.delete_targets))
                sys.exit(2)

        if self.args.remote:
            results = self._delete_remote(self.args.delete_targets)
        else:
            results = self._delete_local(self.args.delete_targets, self.args.force)
        return results

    def prune_images(self):
        """
        Remove dangling images from registry
        :return: 0 if all images deleted or no dangling images found
        """
        self.syscontainers.prune_ostree_images()

        results = self.d.images(filters={"dangling":True}, quiet=True)
        if len(results) == 0:
            return 0

        for img in results:
            self.d.remove_image(img, force=True)
            util.write_out("Removed dangling Image {}".format(img))
        return 0

    def _delete_remote(self, targets):
        results = 0
        for target in targets:
            # _convert_to_skopeo requires registry v1 support while delete requires v2 support
            # args, img = self.syscontainers._convert_to_skopeo(target)

            args = []
            if "http:" in target:
                args.append("--insecure")

            for i in ["oci:", "http:", "https:"]:
                img = target.replace(i, "docker:")

            if not img.startswith("docker:"):
                img = "docker://" + img

            try:
                util.skopeo_delete(img, args)
                util.write_out("Image {} marked for deletion".format(img))
            except ValueError as e:
                util.write_err("Failed to mark Image {} for deletion: {}".format(img, e))
                results = 2
        return results

    def _delete_local(self, targets, force=False):
        results = 0
        for target in targets:
            if not self.args.storage:
                if self.is_duplicate_image(target):
                    util.write_err("Failed to delete Image {}: has duplicate naming; please specify "
                                   "--storage to delete from a specific storage.".format(target))
                    results = 2
                else:
                    if self.syscontainers.has_image(target):
                        self.syscontainers.delete_image(target)
                    else:
                        try:
                            self.d.remove_image(target, force=force)
                        except NotFound as e:
                            util.write_err("Failed to delete Image {}: {}".format(target, e))
                            results = 2
                        except APIError as e:
                            util.write_err("Failed operation for delete Image {}: {}".format(target, e))
                            results = 2

            elif self.args.storage.lower() == "ostree":
                if not self.syscontainers.has_image(target):
                    util.write_err("Failed to delete Image {}: does not exist in ostree.".format(target))
                self.syscontainers.delete_image(target)

            elif self.args.storage.lower() == "docker":
                try:
                    self.d.remove_image(target, force=force)
                except NotFound as e:
                    util.write_err("Failed to delete Image {}: {}".format(target, e))
                    results = 2
                except APIError as e:
                    util.write_err("Failed operation for delete Image {}: {}".format(target, e))
                    results = 2

            else:
                util.write_err("{} is not a valid storage".format(self.args.storage))
                results = 2

        return results



try:
    from . import Atomic
except ImportError:
    from atomic import Atomic  # pylint: disable=relative-import
from .trust import Trust
from .util import skopeo_copy, get_atomic_config, Decompose, write_out, strip_port, is_insecure_registry

ATOMIC_CONFIG = get_atomic_config()


def cli(subparser):
    # atomic pull
    storage = ATOMIC_CONFIG.get('default_storage', "docker")
    pullp = subparser.add_parser("pull", help=_("pull latest image from a repository"),
                                 epilog="pull the latest specified image from a repository.")
    pullp.set_defaults(_class=Pull, func='pull_image')
    pullp.add_argument("--storage", dest="storage", default=storage,
                       help=_("Specify the storage. Default is currently '%s'.  You can"
                              " change the default by editing /etc/atomic.conf and changing"
                              " the 'default_storage' field." % storage))
    pullp.add_argument("-t", "--type", dest="reg_type", default=None,
                       help=_("Pull from an alternative registry type."))
    pullp.add_argument("image", help=_("image id"))


class Pull(Atomic):
    def __init__(self, policy_filename=None):
        """
        :param policy_filename: override policy filename
        """
        super(Pull, self).__init__()
        self.policy_filename=policy_filename

    def pull_docker_image(self):
        self.ping()
        # Add this when atomic registry is incorporated.
        # if self.args.reg_type == "atomic":
        #     pull_uri = 'atomic:'
        # else:
        #     pull_uri = 'docker://'
        fq_name = self.get_fq_image_name(self.args.image)
        registry, _, _, tag, _ = Decompose(fq_name).all
        image = "docker-daemon:{}".format(self.args.image)
        if not self.args.image.endswith(tag):
            image += ":{}".format(tag)
        insecure = True if is_insecure_registry(self.d.info()['RegistryConfig'], strip_port(registry)) else False
        trust = Trust()
        trust.set_args(self.args)
        trust.discover_sigstore(fq_name)
        write_out("Pulling {} ...".format(fq_name))
        skopeo_copy("docker://{}".format(fq_name), image,
                    debug=self.args.debug, insecure=insecure,
                    policy_filename=self.policy_filename)

    def pull_image(self):
        handlers = {
            "ostree" : self.syscontainers.pull_image,
            "docker" : self.pull_docker_image
        }

        handler = handlers.get(self.args.storage)
        if handler is None:
            raise ValueError("Destination not known, please choose --storage=%s" % "|".join(handlers.keys()))
        write_out("Image %s is being pulled to %s ..." % (self.args.image, self.args.storage))
        handler()



from . import util
import argparse

try:
    from . import Atomic
except ImportError:
    from atomic import Atomic  # pylint: disable=relative-import

from .atomic import AtomicError
from docker.errors import NotFound
import requests.exceptions

def cli(subparser, hidden=False):
    # atomic info
    if hidden:
        infop = subparser.add_parser("info", argument_default=argparse.SUPPRESS)

    else:
        infop = subparser.add_parser(
            "info", help=_("display label information about an image"),
            epilog="atomic info attempts to read and display the LABEL "
            "information about an image")

    infop.set_defaults(_class=Info, func='info_tty')
    infop.add_argument("--remote", dest="force",
                       action='store_true', default=False,
                       help=_('ignore local images and only scan registries'))
    infop.add_argument("--storage", default="", dest="storage",
                       help=_("Specify the storage of the image. "
                              "If not specified and there are images with the same name in "
                              "different storages, you will be prompted to specify."))
    infop.add_argument("image", help=_("container image"))


def cli_version(subparser, hidden=False):
    if hidden:
        versionp = subparser.add_parser("version", argument_default=argparse.SUPPRESS)

    else:
        versionp = subparser.add_parser(
            "version", help=_("display image 'Name Version Release' label"),
            epilog="atomic version displays the image version information, if "
            "it is provided")
    versionp.add_argument("-r", "--recurse", default=False, dest="recurse",
                          action="store_true",
                          help=_("recurse through all layers"))
    versionp.add_argument("--storage", default="", dest="storage",
                          help=_("Specify the storage of the image. "
                                 "If not specified and there are images with the same name in "
                                 "different storages, you will be prompted to specify."))
    versionp.set_defaults(_class=Info, func='print_version')
    versionp.add_argument("image", help=_("container image"))


class Info(Atomic):
    def __init__(self):
        super(Info, self).__init__()

    def version(self):
        if not self.args.storage:
            if self.is_duplicate_image(self.image):
                raise ValueError("Found more than one Image with name {}; "
                                 "please specify with --storage.".format(self.image))
            else:
                if self.syscontainers.has_image(self.image):
                    return self.syscontainers.version(self.image)
                try:
                    self.d.inspect_image(self.image)
                except (NotFound, requests.exceptions.ConnectionError):
                    self._no_such_image()

        elif self.args.storage.lower() == "ostree":
            if self.syscontainers.has_image(self.image):
                return self.syscontainers.version(self.image)
            self._no_such_image()

        elif self.args.storage.lower() == "docker":
            try:
                self.d.inspect_image(self.image)
            except (NotFound, requests.exceptions.ConnectionError):
                self._no_such_image()

        else:
            raise ValueError("{} is not a valid storage".format(self.args.storage))

        if self.args.recurse:
            return self.get_layers()
        else:
            return [self._get_layer(self.image)]

    def get_version(self):
        versions = []
        for layer in self.version():
            version = "None"
            if "Version" in layer and layer["Version"] != '':
                version = layer["Version"]
            versions.append({"Image": layer['RepoTags'], "Version": version, "iid": layer['Id']})
        return versions

    def info_tty(self):
        util.write_out(self.info())

    def info(self):
        """
        Retrieve and print all LABEL information for a given image.
        """
        buf = ""

        def _no_label():
            return ""
        if not self.args.storage:
            if self.is_duplicate_image(self.image):
                raise ValueError("Found more than one Image with name {}; "
                                 "please specify with --storage.".format(self.image))

            if self.syscontainers.has_image(self.image):
                if not self.args.force:
                    buf += ("Image Name: {}".format(self.image))
                    manifest = self.syscontainers.inspect_system_image(self.image)
                    labels = manifest["Labels"]
                    for label in labels:
                        buf += ('\n{0}: {1}'.format(label, labels[label]))
                    return buf
            # Check if the input is an image id associated with more than one
            # repotag.  If so, error out.
            else:
                try:
                    iid = self._is_image(self.image)
                    self.image = self.get_fq_name(self._inspect_image(iid))
                except AtomicError:
                    if self.args.force:
                        self.image = util.find_remote_image(self.d, self.image)
                    if self.image is None:
                        self._no_such_image()

        elif self.args.storage.lower() == "ostree":
            if self.syscontainers.has_image(self.image):
                if not self.args.force:
                    buf += ("Image Name: {}".format(self.image))
                    manifest = self.syscontainers.inspect_system_image(self.image)
                    labels = manifest["Labels"]
                    for label in labels:
                        buf += ('\n{0}: {1}'.format(label, labels[label]))
                    return buf
            else:
                self._no_such_image()

        elif self.args.storage.lower() == "docker":
            if self.is_iid():
                self.get_fq_name(self._inspect_image())
            # The input is not an image id
            else:
                try:
                    iid = self._is_image(self.image)
                    self.image = self.get_fq_name(self._inspect_image(iid))
                except AtomicError:
                    if self.args.force:
                        self.image = util.find_remote_image(self.d, self.image)
                    if self.image is None:
                        self._no_such_image()

        else:
            raise ValueError("{} is not a valid storage".format(self.args.storage))

        buf += ("Image Name: {}".format(self.image))
        inspection = None
        if not self.args.force:
            inspection = self._inspect_image(self.image)
            # No such image locally, but fall back to remote
        if inspection is None:
            # Shut up pylint in case we're on a machine with upstream
            # docker-py, which lacks the remote keyword arg.
            # pylint: disable=unexpected-keyword-arg
            inspection = util.skopeo_inspect("docker://" + self.image)
            # image does not exist on any configured registry
        if 'Config' in inspection and 'Labels' in inspection['Config']:
            labels = inspection['Config']['Labels']
        elif 'Labels' in inspection:
            labels = inspection['Labels']
        else:
            _no_label()

        if labels is not None and len(labels) is not 0:
            for label in labels:
                buf += ('\n{0}: {1}'.format(label, labels[label]))
        else:
            _no_label()
        return buf

    def print_version(self):
        versions = self.get_version()
        max_version_len = len(max([x['Version'] for x in versions], key=len)) + 2
        max_version_len = max_version_len if max_version_len > 9 else 9
        max_img_len = len(max([y for x in versions for y in x['Image']], key=len)) + 9
        max_img_len = max_img_len if max_img_len > 12 else 12
        col_out = "{0:" + str(max_img_len) + "} {1:" + str(max_version_len) + "} {2:10}"
        util.write_out("")
        util.write_out(col_out.format("IMAGE NAME", "VERSION", "IMAGE ID"))
        for layer in versions:
            for int_img_name in range(len(layer['Image'])):
                version = layer['Version'] if int_img_name < 1 else ""
                iid = layer['iid'][:12] if int_img_name < 1 else ""
                space = "" if int_img_name < 1 else "  Tag: "
                util.write_out(col_out.format(space + layer['Image'][int_img_name], version, iid))
        util.write_out("")




import argparse
from . import util
from .util import add_opt
from .install import INSTALL_ARGS
from docker.errors import NotFound

try:
    from . import Atomic
except ImportError:
    from atomic import Atomic # pylint: disable=relative-import

def cli(subparser):
    # atomic uninstall
    uninstallp = subparser.add_parser(
        "uninstall", help=_("execute container image uninstall method"),
        epilog="atomic uninstall attempts to read the LABEL UNINSTALL "
        "field in the image, if it does not exist atomic will "
        "remove the image from your machine.  You could add a "
        "LABEL UNINSTALL command to your Dockerfile like: 'LABEL "
        "UNINSTALL %s'" % Uninstall.print_uninstall())
    uninstallp.set_defaults(_class=Uninstall, func='uninstall')
    add_opt(uninstallp)
    uninstallp.add_argument("-n", "--name", dest="name", default=None,
                            help=_("name of container"))
    uninstallp.add_argument("-f", "--force", default=False, dest="force",
                            action="store_true",
                            help=_("remove all containers based on this "
                                   "image"))
    uninstallp.add_argument("image", help=_("container image"))
    uninstallp.add_argument("args", nargs=argparse.REMAINDER,
                            help=_("Additional arguments appended to the "
                                   "image uninstall method"))

class Uninstall(Atomic):
    def __init__(self):
        super(Uninstall, self).__init__()

    def uninstall(self):
        if self.syscontainers.get_checkout(self.args.image):
            return self.syscontainers.uninstall(self.args.image)

        self.inspect = self._inspect_container()
        if self.inspect and self.force:
            self.force_delete_containers()
        try:
            # Attempt to remove container, if it exists just return
            self.d.stop(self.name)
            self.d.remove_container(self.name)
        except NotFound:
            # On exception attempt to remove image
            pass

        self.inspect = self._inspect_image()
        if not self.inspect:
            raise ValueError("Image '%s' is not installed" % self.image)

        args = self._get_args("UNINSTALL")
        if args:
            cmd = self.gen_cmd(args + self.quote(self.args.args))
            cmd = self.sub_env_strings(cmd)
            self.display(cmd)
            util.check_call(cmd, env=self.cmd_env())

        if self.name == self.image:
            util.write_out("docker rmi %s" % self.image)
            util.check_call([self.docker_binary(), "rmi", self.image])

    @staticmethod
    def print_uninstall():
        return "%s %s %s" % (util.default_docker(), " ".join(INSTALL_ARGS), "/usr/bin/UNINSTALLCMD")




import requests
import json
import os
from base64 import b64decode
from . import util
import hashlib


try:
    from urllib.request import parse_http_list, parse_keqv_list
except ImportError:
    from urllib2 import parse_http_list, parse_keqv_list #pylint: disable=import-error


def jose_base64_urldecode(in_str):
    in_str = in_str.replace("\n", "")
    in_str = in_str.replace(" ", "")
    str_len = len(in_str) % 4
    if str_len not in [0, 2, 3]:
        raise ValueError("{} is an illegal base64url string".format(in_str))
    if str_len == 2:
        in_str += "=="
    elif str_len == 3:
        in_str += "="
    return b64decode(in_str)


class RegistryConnection():

    def __init__(self, debug):
        self.headers = {
            'Accept': '[application/vnd.oci.image.manifest.v1+json,'
                      'application/vnd.docker.distribution.manifest.v2+json,'
                      'application/vnd.docker.distribution.manifest.v1+prettyjws,'
                      'application/vnd.docker.distribution.manifest.v1+json]'
        }
        self.auth_headers = {}
        self.hostname = None
        self.scheme = None
        self.needs_auth = False
        self.token = None
        self.token_realm = None
        self.token_service = None
        self.token_scope = None
        self._debug = debug
        self.orig_manifest = None
        self.manifest_json = None
        self.whatever_json = None
        self.tag_json = None
        self.registry = None
        self.local_tokens = None
        self.schema_version = None
        self.pinged = False
        self.verify = True
        self.port = None

    def inspect_schema1(self, name, tag):
        v1compat = json.loads(self.manifest_json['history'][0]['v1Compatibility'])
        return {
            'Name': name,
            'Tag': tag,
            'Digest': "sha256:{}".format(self.get_digest()),
            'RepoTags':  self.tag_json['tags'],
            'Created': v1compat['created'],
            'DockerVersion': v1compat['docker_version'],
            'Labels': v1compat['config']['Labels'],
            'Architecture': self.manifest_json['architecture'],
            'Os': v1compat['os'],
            'Layers': [x['blobSum'] for x in self.manifest_json['fsLayers']][::-1]
            }

    def inspect_schema2(self, name, tag):
        return {
            'Name': name,
            'Tag': tag,
            'Digest': "sha256:{}".format(self.get_digest()),
            'RepoTags':  self.tag_json['tags'],
            'Created':  self.whatever_json['created'],
            'DockerVersion': self.whatever_json['docker_version'],
            'Labels':  self.whatever_json['config']['Labels'],
            'Architecture': self.whatever_json['architecture'],
            'Os': self.whatever_json['os'],
            'Layers': [x['digest'] for x in self.manifest_json['layers']][::-1]
        }

    def get_digest(self):
        if 'signatures' in self.manifest_json:
            return self._get_digest_from_signature()
        else:
            return hashlib.sha256(self.orig_manifest.encode('utf-8')).hexdigest()

    def _get_digest_from_signature(self):
        for i in self.manifest_json['signatures']:
            _protected = i['protected']
            protected_js = json.loads(jose_base64_urldecode(_protected).decode('utf-8'))
            format_length = protected_js['formatLength']
            format_tail = jose_base64_urldecode(protected_js['formatTail']).decode('utf-8')
            orig = [ord(x) for x in self.orig_manifest]
            formatted = orig[:format_length] + [ord(x) for x in format_tail]
            return hashlib.sha256("".join([chr(x) for x in formatted]).encode('utf-8')).hexdigest()

    @staticmethod
    def load_local_tokens():
        tokens = {}
        token_file_name = os.path.expanduser('~/.docker/config.json')
        if not os.path.exists(token_file_name):
            return {}
        with open(token_file_name) as token_file:
            token_data = json.load(token_file)
        for registry in token_data['auths']:
            tokens[registry] = token_data['auths'][registry]['auth']
        return tokens

    def get(self, url, skip_auth=False):
        if self.needs_auth and not skip_auth:
            if not self.token:
                self.get_token()
            headers = self.get_auth_headers()
        else:
            headers = self.headers

        self.debug("GET_URL: %s" % url)
        self.debug("GET_HEADER: %s" % headers)
        self.debug("GET_VERIFY: %s" % self.verify)
        results = requests.get(url, headers=headers, verify=self.verify)
        if results.reason == "Unauthorized" and self.needs_auth:
            # Auth registries dont tell us if credentials are bad or the GET URL is bad
            raise RegistryAuthError("Unable to authenticate to {} or repository, image, "
                                    "or tags names are bad".format(self.registry))
        return results

    def _create_self_auth_headers(self):
        headers = self.headers.copy()
        headers['Authorization'] = "Bearer {}".format(self.token)
        self.auth_headers = headers

    def get_auth_headers(self):
        if len(self.auth_headers) is 0:
            self._create_self_auth_headers()
        return self.auth_headers

    def debug(self, msg):
        if self._debug:
            util.write_out(msg)

    @staticmethod
    def get_components_from_bearer(header):
        _, _, value = header.partition("Bearer")
        opts = parse_keqv_list(parse_http_list(value)) if value else None
        return opts.get('realm', None), opts.get('service', None), opts.get('scope', None)

    def set_token_from_header(self, bearer):
        realm, service, scope = self.get_components_from_bearer(bearer)
        self._set_token(realm, service, scope)

    def _set_token(self, realm, service, scope):
        self.token_realm = realm
        self.token_service = service
        self.token_scope = scope

    def get_token(self):
        local_tokens = self.local_tokens if self.local_tokens else self.load_local_tokens()
        if self.token_scope and self.token_service:
            url = '{}?service={}&scope={}'.format(self.token_realm, self.token_service, self.token_scope)
            self.token = self.get(url, skip_auth=True).json()['token']
        elif not self.token_scope and not self.token_service and self.hostname in local_tokens:
            host_token = (local_tokens[self.hostname]).encode()
            self.token = str(b64decode(host_token)).split(':')[-1]
        else:
            raise RegistryAuthError("Failed to obtain token")
        self.debug("Set token to {}".format(self.token))

    def set_token_scope(self, repo, image):
        self.token_scope = "repository:{}/{}:pull".format(repo, image)
        self.debug("Set token_scope to {}".format(self.token_scope))

    def hostname_has_port(self):
        if len(self.hostname.split(':')) > 1:
            return True
        return False


class RegistryInspectError(Exception):
    pass


class RegistryAuthError(Exception):
    pass


class RegistryInspect():

    def __init__(self, registry=None, repo=None, image=None, tag=None, orig_input=None, debug=False):
        self.debug = debug
        self.registries = util.get_registries()
        self.registry = registry
        self.repo = repo
        self.image = image
        self.tag = tag
        self.orig_input = orig_input
        if self.debug:
            util.output_json(self.registries)
        self.rc = RegistryConnection(debug=self.debug)
        self._setup_rc()

    def _setup_rc(self):
        if self.registry:
            try:
                self.rc.hostname = (x['hostname'] for x in self.registries if x['name'] == self.registry).__next__()
                self.rc.verify = (x['secure'] for x in self.registries if x['name'] == self.registry).__next__()
            except AttributeError:
                self.rc.hostname = (x['hostname'] for x in self.registries if x['name'] == self.registry).next()  # pylint: disable=next-method-called
                self.rc.verify = (x['secure'] for x in self.registries if x['name'] == self.registry).next()  # pylint: disable=next-method-called
            self.rc.registry = self.registry
            self.rc.needs_auth = False

    def inspect(self):
        def _get_inspect_info():

            if self.rc.schema_version == 1:
                return self.rc.inspect_schema1(self.assemble_fqdn(include_tag=False), self.tag)
            else:
                return self.rc.inspect_schema2(self.assemble_fqdn(include_tag=False), self.tag)

        if not self.registry:
            self.find_image_on_registry()
        if not self.rc.pinged:
            self.ping()
        if not self.rc.manifest_json:
            self.get_manifest()
        self.rc.schema_version = self.rc.manifest_json['schemaVersion']
        if self.rc.schema_version == 2:
            self.get_blob_info()
        self.get_tag_list()
        return _get_inspect_info()



    def assemble_fqdn(self, include_tag=True):
        fqdn = "{}".format(self.registry)
        fqdn = fqdn if not self.repo else "{}/{}".format(fqdn, self.repo)
        fqdn += "/{}".format(self.image)
        if include_tag:
            fqdn += ":{}".format(self.tag)
        return fqdn



    def find_image_on_registry(self):
        """
        Find the fully qualified image name for given input when
        registry is unknown
        :return: String fqdn
        """
        for i in [x for x in self.registries if x['search']]:
            docker_repo = False
            self.registry = i['name']
            if not self.repo and self.registry == "docker.io":
                docker_repo = True
                self.repo = 'library'
                self.rc = RegistryConnection(debug=self.debug)
            self._setup_rc()
            try:
                util.write_out("Trying {}".format(self.assemble_fqdn(include_tag=True)))
                self.ping()
            except RegistryInspectError as e:
                util.write_out(str(e))
                continue
            except RegistryAuthError:
                pass

            try:
                manifest_json = self.get_manifest()
                self.rc.manifest_json = manifest_json
                fq = self.registry
                if self.repo:
                    fq += "/{}".format(self.repo)
                fq += "/{}:{}".format(self.image, self.tag)
                return fq

            except RegistryAuthError:
                self.rc = RegistryConnection(debug=self.debug)
            except RegistryInspectError:
                self.rc = RegistryConnection(debug=self.debug)
            if docker_repo:
                self.repo = None
        raise RegistryInspectError("Unable to resolve {}".format(self.orig_input))

    def _ping(self, _scheme, port=None):
        """
        :param _scheme: str(http|https_
        :param port: optional port
        :return: 0 for sucess, 1 for SSL failure, 2 for connection error
        """
        # Inject the port if provided
        hostname = self.rc.hostname if not port else "{}:{}".format(self.rc.hostname, port)
        url = '{}://{}/v2/'.format(_scheme, hostname)
        if self.debug:
            util.write_out("URL: {}".format(url))
        try:
            results = self.rc.get(url)
            if results.reason == "Unauthorized" and self.rc.token:
                raise RegistryAuthError("Unable to establish authenticated connection to {}".format(self.registry))
            if results.reason == "Unauthorized":  # Need a token?
                # Auth hasn't been set up; do so now
                self.rc.needs_auth = True
                self.rc.set_token_from_header(results.headers['Www-authenticate'])
                if not self.rc.token_scope and self.rc.token_service:
                    self.rc.set_token_scope(self.repo, self.image)
                return 3
        except requests.exceptions.SSLError as e:
            if self.debug:
                util.write_out(str(e))
            return 1
        except requests.exceptions.ConnectionError as e:
            if self.debug:
                util.write_out(str(e))
            return 2
        try:
            # Sometimes you get a result from an actual page. A ping response
            # should be an empty dict
            if bool(results.json()):
                return 4
        except ValueError:
            return 4

        if self.rc.hostname != hostname:
            self.rc.port = 5000
        return 0

    def ping(self):
        def _set_auth():
            # If the reg needs auth, attempt again with creds
            self.rc.scheme = scheme
            self.rc.pinged = True
            if self.rc.needs_auth:
                self._ping(scheme)

        scheme = 'https'
        rc = self._ping(scheme)
        if rc == 0:
            return _set_auth()
        if rc in [2, 4] and not self.rc.hostname_has_port():
            if self._ping(scheme, port=5000) == 0:
                return _set_auth()
        else:
            # Now do http
            scheme = 'http'
            rc = self._ping(scheme)
            if rc == 0:
                return _set_auth()
            if rc in [2, 4] and not self.rc.hostname_has_port():
                if self._ping(scheme, port=5000) == 0:
                    return _set_auth()
        raise RegistryInspectError("Unable to connect to registry '{}'".format(self.rc.registry))

    def _assemble_hostname(self):
        return self.rc.hostname if not self.rc.port else "{}:{}".format(self.rc.hostname, self.rc.port)

    def get_manifest(self):
        if self.rc.manifest_json:
            return self.rc.manifest_json

        url = '{}://{}/v2/{}/manifests/{}'.format(self.rc.scheme, self._assemble_hostname(),
                                                  os.path.join(*[x for x in [self.repo, self.image] if x]), self.tag)
        results = self.rc.get(url)
        if results.status_code == 200:
            self.rc.manifest_json = results.json()
            self.rc.orig_manifest = results.content.decode('utf-8')
            return self.rc.manifest_json
        else:
            raise RegistryInspectError("Unable to obtain manifest for "
                                       "{}/{}/{}:{}".format(self.registry, self.repo, self.image, self.tag))

    def get_blob_info(self):
        def _get_digest_from_json():
            if self.rc.manifest_json and self.rc.manifest_json.get('config', None):
                digest = self.rc.manifest_json.get('config').get('digest', None)
                if digest:
                    return digest
            raise RegistryInspectError("You must instantiate the image's manifest date to obtain its digest")

        url = '{}://{}/v2/{}/{}/blobs/{}'.format(self.rc.scheme,
                                                 self._assemble_hostname(),
                                                 self.repo,
                                                 self.image,
                                                 _get_digest_from_json())
        results = self.rc.get(url)
        if results.status_code == 200:
            self.rc.whatever_json = results.json()
            return self.rc.whatever_json
        else:
            raise RegistryInspectError("Unable to obtain blob for "
                                       "{}/{}/{}:{}".format(self.registry, self.repo, self.image, self.tag))

    def get_tag_list(self):
        url = '{}://{}/v2/{}/{}/tags/list'.format(self.rc.scheme, self._assemble_hostname(), self.repo, self.image)
        results = self.rc.get(url)
        if results.status_code == 200:
            self.rc.tag_json = results.json()
            return self.rc.tag_json
        else:
            raise RegistryInspectError("Unable to obtain tag information for "
                                       "{}/{}/{}:{}".format(self.registry, self.repo, self.image, self.tag))




import shutil
import selinux
import requests
import os

from . import util
from .Export import export_docker
from .Import import import_docker
from .util import NoDockerDaemon, default_docker_lib

try:
    from subprocess import DEVNULL  # pylint: disable=no-name-in-module
except ImportError:
    DEVNULL = open(os.devnull, 'wb')

try:
    from . import Atomic
except ImportError:
    from atomic import Atomic # pylint: disable=relative-import

def cli(subparser):
    # atomic storage
    storagep = subparser.add_parser(
        "storage", aliases=['migrate'], help=_("manage container storage"),
        epilog="atomic storage command allows you to setup/reset "
        "container storage")
    storage_subparser = storagep.add_subparsers(help=_("storage commands"))
    # atomic storage export
    exportp = storage_subparser.add_parser("export",
                                           help=_("export containers and associated contents into a filesystem directory"),
                                           epilog="Export containers. "
                                           "The export command exports images, "
                                           "containers, and volumes into a filesystem directory.")
    exportp.set_defaults(_class=Storage, func='Export')
    exportp.add_argument("--graph", dest="graph",
                         default=default_docker_lib(),
                         help=_("Root of the Docker runtime (Default: %s)" % default_docker_lib()))
    exportp.add_argument("--dir", dest="export_location",
                         default="/var/lib/atomic/migrate",
                         help=_("Path for exporting container's content (Default: /var/lib/atomic/migrate)"))

    # atomic storage import
    importp = storage_subparser.add_parser("import", help=_("import containers associated contents from a filesystem directory"),
                                           epilog="Import containers. "
                                           "The import command imports images,"
                                           "containers, and volumes from a filesystem directory.")
    importp.set_defaults(_class=Storage, func='Import')
    importp.add_argument("--graph", dest="graph",
                         default=default_docker_lib(),
                         help=_("Root of the Docker runtime (Default: %s)" % default_docker_lib()))

    importp.add_argument("--dir", dest="import_location",
                         default="/var/lib/atomic/migrate",
                         help=_("Path for importing container's content (Default: /var/lib/atomic/migrate)"))

    # atomic storage modify
    modifyp = storage_subparser.add_parser("modify",help='modify default storage setup')
    modifyp.add_argument('--add-device', metavar="DEVICE", dest="devices", default=[], action='append',
                         help=_("add block devices to storage pool"))
    modifyp.add_argument('--remove-device', metavar="DEVICE", dest="remove_devices", default=[], action='append',
                         help=_("remove block devices from storage pool"))
    modifyp.add_argument('--remove-unused-devices', action='store_true',
                         help=_("remove all unused block devices from storage pool"))
    modifyp.add_argument('--driver', dest="driver", default=None, help='The storage backend driver', choices=['devicemapper', 'overlay', 'overlay2'])
    modifyp.add_argument('--vgroup', dest="vgroup", default=None, help='The storage volume group')
    modifyp.set_defaults(_class=Storage, func='modify')

    # atomic storage reset
    resetp = storage_subparser.add_parser("reset",
                                          help=_("delete all containers/images from your system. Reset storage to its initial configuration."))
    resetp.set_defaults(_class=Storage, func='reset')

def query_pvs(pv, fields):
    return util.check_output([ "pvs", "--noheadings", "-o",  fields, "--unit", "b", pv ]).split()

def list_pvs(vgroup):
    res = [ ]
    if vgroup:
        for l in util.check_output([ "pvs", "--noheadings", "-o",  "vg_name,pv_name" ]).splitlines():
            fields = l.split()
            if len(fields) == 2 and fields[0] == vgroup:
                res.append(fields[1])
    return res

def list_lvs(vgroup):
    if vgroup:
        return map(lambda s: s.strip(), # pylint: disable=deprecated-lambda, map-builtin-not-iterating
                   util.check_output([ "lvs", "--noheadings", "-o", "name", vgroup ]).splitlines())
    else:
        return [ ]

def list_parents(dev):
    return util.check_output([ "lsblk", "-snlp", "-o", "NAME", dev ]).splitlines()[1:]

def list_children(dev):
    return util.check_output([ "lsblk", "-nlp", "-o", "NAME", dev ]).splitlines()[1:]

def get_dss_vgroup(conf):
    vgroup = util.sh_get_var_in_file(conf, "VG", "")
    if vgroup == "":
        for l in open("/proc/mounts", "r").readlines():
            fields = l.split()
            if fields[1] == "/" and fields[0].startswith("/dev"):
                vgroup = util.check_output([ "lvs", "--noheadings", "-o",  "vg_name", fields[0]]).strip()
    return vgroup

def get_dss_devs(conf):
    return util.sh_get_var_in_file(conf, "DEVS", "").split()

class Storage(Atomic):
    dss_conf = "/etc/sysconfig/docker-storage-setup"
    dss_conf_bak = dss_conf + ".bkp"

    def reset(self):
        root = "/var/lib/docker"
        try:
            self.d.info()
            raise ValueError("Docker daemon must be stopped before resetting storage")
        except requests.exceptions.ConnectionError:
            pass

        util.check_call(["docker-storage-setup", "--reset"], stdout=DEVNULL)
        util.call(["umount", root + "/devicemapper"], stderr=DEVNULL)
        util.call(["umount", root + "/overlay"], stderr=DEVNULL)
        util.call(["umount", root + "/overlay2"], stderr=DEVNULL)
        shutil.rmtree(root)
        os.mkdir(root)
        try:
            selinux.restorecon(root.encode("utf-8"))
        except (TypeError, OSError):
            selinux.restorecon(root)

    def modify(self):
        try:
            shutil.copyfile(self.dss_conf, self.dss_conf_bak)
            if len(self.args.remove_devices) > 0:
                self._remove_devices(self.args.remove_devices, only_unused=False)
            if self.args.remove_unused_devices:
                self._remove_devices(get_dss_devs(self.dss_conf), only_unused=True)
            if self.args.driver:
                self._driver(self.args.driver)
            if self.args.vgroup is not None:
                self._vgroup(self.args.vgroup)
            if len(self.args.devices) > 0:
                self._add_device(self.args.devices)
            if util.call(["docker-storage-setup"]) != 0:
                os.rename(self.dss_conf_bak, self.dss_conf)
                util.call(["docker-storage-setup"])
                raise ValueError("docker-storage-setup failed")
        except:
            if os.path.exists(self.dss_conf_bak):
                os.rename(self.dss_conf_bak, self.dss_conf)
            raise
        finally:
            if os.path.exists(self.dss_conf_bak):
                os.remove(self.dss_conf_bak)

    def _add_device(self, devices):
        util.sh_modify_var_in_file(self.dss_conf, "DEVS",
                                   lambda old: util.sh_set_add(old, devices))

    def _remove_devices(self, devices, only_unused):
        vgroup = get_dss_vgroup(self.dss_conf)
        pvs = list_pvs(vgroup)
        lvs = list_lvs(vgroup)
        n_pvs = len(pvs)
        devices = set(devices)
        parents = None
        for pv in pvs:
            parents = list_parents(pv)
            if set(parents).isdisjoint(devices):
                continue
            devices -= set(parents)

            if query_pvs(pv, "pv_used")[0][:-1] != '0':
                if only_unused:
                    continue
                else:
                    util.check_call([ "pvmove", pv ])

            if n_pvs > 1:
                util.check_call([ "vgreduce", vgroup, pv ])
            elif len(lvs) == 0:
                util.check_call([ "vgremove", vgroup ])
            n_pvs -= 1
            util.check_call([ "wipefs", "-a", pv ])
            util.sh_modify_var_in_file(self.dss_conf, "DEVS",
                                       lambda old: util.sh_set_del(old, parents))
            if len(parents) == 1:
                children = list_children(parents[0])
                if len(children) == 1 and children[0] == pv:
                    util.check_call([ "wipefs", "-a", parents[0] ])
        if len(devices) > 0:
            raise ValueError("Not part of the storage pool: {}".format(", ".join(devices)))

    def _driver(self, driver):
        util.sh_modify_var_in_file(self.dss_conf, "STORAGE_DRIVER",
                                   lambda old: driver)

    def _vgroup(self, vgroup):
        util.sh_modify_var_in_file(self.dss_conf, "VG",
                                   lambda old: vgroup)

    def Export(self):
        try:
            export_docker(self.args.graph, self.args.export_location, self.args.assumeyes)
        except requests.exceptions.ConnectionError:
            raise NoDockerDaemon()

    def Import(self):
        self.ping()
        try:
            import_docker(self.args.graph, self.args.import_location, self.args.assumeyes)
        except requests.exceptions.ConnectionError:
            raise NoDockerDaemon()



from . import Atomic
from . import util
import argparse
import json as Json
import tty
import sys
import termios
import select
from os import isatty
from operator import itemgetter
import requests
from .util import NoDockerDaemon

def check_negative(value):
    ivalue = int(value)
    if ivalue < 1:
        raise argparse.ArgumentTypeError("%s must be a positive integer value greater than 0." % value)
    return ivalue

def cli(subparser):
    # atomic top
    topp = subparser.add_parser("top",
                                help=_("Show top-like stats about processes running in containers"))
    topp.set_defaults(_class=Top, func='atomic_top')
    topp.add_argument("-d", type=int, default=1, help=_("Interval (secs) to refresh process information"))
    topp.add_argument("-o", "--optional", help=_("Additional fields to display"), nargs='?', choices=['time', 'stime', 'ppid', 'uid', 'gid', 'user', 'group'])
    topp.add_argument("-n", help=_("Number of iterations"), type=check_negative)
    topp.add_argument("containers", nargs="*", help=_("list of containers to monitor, leave blank for all"))


class Top(Atomic):
    """
    Class to support atomic top; based on the Atomic class
    """
    # Set to true to output debug information

    def __init__(self):
        super(Top, self).__init__()
        self.input_var = None
        self._sort = 'CID'
        self.name_id = {}
        self.optional = None
        self.titles = None
        self.debug = False
        # To add a new column to the output, create a new dict inside self.headers
        # The fields are as follows:
        # shortname: <str> a unique key used to describe the dict
        # column: <str> column title, use parans to define a sort character
        # character: <str> a single character the user presses to sort on that column
        # sort: <bool> to determine IF the column can be sorted upon
        # index: <int> order the columns display left to right. Do not change 0-5
        # _field: <str> If the width of the column needs to be dynamically sized, put
        #               the formatting in _field, set field to None, and be sure to
        #               add a key for _min_width
        # field: <str> the preferred format for the column
        # active: if the column should be active by default
        # ps_opt: <str> the descriptor for ps to get the column's data (ps -o)
        # sort_order: <bool> True reverses the sort order from ascending to descending
        # _min_width: <int> Minimum field width if you are using dynamic fields
        self.headers = [
            {'shortname': '%CPU', 'column': '(C)PU', 'character': 'c', 'sort': True, 'index': 3,
             'field': '{:6}', 'active': True, 'ps_opt': 'pcpu', 'sort_order': True},
            {'shortname': 'CID', 'column': 'CONTA(I)NER', 'character': 'i', 'sort': True, 'index': 0,
             'field': '{:12.12}', 'active': True, 'ps_opt': None, 'sort_order': False},
            {'shortname': 'NAME', 'column': '(N)AME', 'character': 'n', 'sort': True, 'index': 1,
             '_field': '{:WIDTH}', 'field': None, 'active': True, 'ps_opt': None, 'sort_order': False,
             '_min_width': 10},
            {'shortname': 'PID', 'column': '(P)ID', 'character': 'p', 'sort': True, 'index': 2,
             'field': '{:<10}', 'active': True, 'ps_opt': 'pid', 'sort_order': False},
            {'shortname': '%MEM', 'column': '(M)EM', 'character': 'm', 'sort': True, 'index': 4,
             'field': '{:6}', 'active': True, 'ps_opt': 'pmem', 'sort_order': True},
            {'shortname': 'CMD', 'column': 'CMD', 'character': None, 'sort': False, 'index': 99,
             '_field': '{:WIDTH}', 'field': None, 'active': True, 'ps_opt': 'cmd', '_min_width': 10},
            {'shortname': 'TIME', 'column': '(T)IME', 'character': 't', 'sort': True, 'index': 6,
             'field': '{:10}', 'active': False, 'ps_opt': 'time', 'sort_order': True},
            {'shortname': 'STIME', 'column': '(S)TIME', 'character': 's', 'sort': True, 'index': 7,
             'field': '{:10}', 'active': False, 'ps_opt': 'stime', 'sort_order': True},
            {'shortname': 'PPID', 'column': 'PPI(D)', 'character': 'd', 'sort': True, 'index': 8,
             'field': '{:10}', 'active': False, 'ps_opt': 'ppid', 'sort_order': False},
            {'shortname': 'UID', 'column': '(U)ID', 'character': 'u', 'sort': True, 'index': 9,
             'field': '{:6}', 'active': True, 'ps_opt': 'uid', 'sort_order': True},
            {'shortname': 'GID', 'column': '(G)ID', 'character': 'g', 'sort': True, 'index': 10,
             'field': '{:6}', 'active': True, 'ps_opt': 'gid', 'sort_order': True},
            {'shortname': 'USER', 'column': 'USER', 'character': None, 'sort': False, 'index': 11,
             'field': '{:10}', 'active': False, 'ps_opt': 'user'},
            {'shortname': 'GROUP', 'column': 'GROUP', 'character': None, 'sort': False, 'index': 12,
             'field': '{:10}', 'active': False, 'ps_opt': 'group'}
        ]

    def _activate_optionals(self):
        """
        Sets the active bool to True for any optional  ps
        arguments.
        :return: None
        """
        if self.args.optional:
            for option in self.args.optional:
                for header in self.headers:
                    if header['ps_opt'] == option:
                        header['active'] = True

    def _set_dynamic_column_widths(self, ps_info):
        for header in self.headers:
            if '_field' in header:
                max_widths = [len(x[header['shortname']]) for x in ps_info]
                max_width = header['_min_width'] if not max_widths else max(max_widths)
                header['field'] = header['_field'].replace('WIDTH', str(max_width))

    def json(self):
        """
        Main sub-function for top
        :return: None
        """
        # Make sure the docker daemon is running
        self.ping()
        # Activate optional columns
        self._activate_optionals()
        proc_info = []
        if len(self.args.containers) < 1:
            try:
                con_ids = [x['Id'] for x in self.get_active_containers(refresh=True)]
            except requests.exceptions.ConnectionError:
                raise NoDockerDaemon()
        else:
            con_ids = []
            self.get_active_containers(refresh=True)
            # verify the inputs are valid
            for user_input in self.args.containers:
                con_ids.append(self.get_input_id(user_input))
        for cid in con_ids:
            proc_info += self.get_pids_by_container(cid)

        return Json.dumps(self.reformat_ps_info(proc_info))

    def atomic_top(self):
        """
        Main sub-function for top
        :return: None
        """
        # Make sure the docker daemon is running
        self.ping()
        # Set debug bool
        self.set_debug()
        # Activate optional columns
        self._activate_optionals()
        # Do we have a tty?
        # Can run ./atomic top <&- to replicate no tty
        has_tty = isatty(0)
        # Set up terminal, input handling
        if has_tty:
            fd = sys.stdin.fileno()
            old_settings = termios.tcgetattr(fd)
        sort_vals = [x['character'].upper() for x in self.headers if x['active'] and x['sort']]
        counter = 0
        while True:
            proc_info = []
            if len(self.args.containers) < 1:
                try:
                    con_ids = [x['Id'] for x in self.get_active_containers(refresh=True)]
                except requests.exceptions.ConnectionError:
                    raise NoDockerDaemon()
            else:
                con_ids = []
                self.get_active_containers(refresh=True)
                # verify the inputs are valid
                for user_input in self.args.containers:
                    con_ids.append(self.get_input_id(user_input))
            for cid in con_ids:
                proc_info += self.get_pids_by_container(cid)
            # Reset screen
            if not self.debug and has_tty:
                util.write_out("\033c")
            sorted_info = self.reformat_ps_info(proc_info)
            self._set_dynamic_column_widths(sorted_info)
            self.output_top(sorted_info)
            if has_tty:
                tty.setraw(sys.stdin.fileno())
            i, _, _ = select.select([sys.stdin], [], [], self.args.d)
            if i and has_tty:
                ch = sys.stdin.read(1)
                # Detect 'q' or Cntrl-c
                if ch.upper() == 'q'.upper() or ord(ch) == 3:
                    # reset terminal
                    termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
                    raise SystemExit
                # Detect of the character pushed is one in self.headers
                elif ch.upper() in sort_vals:
                    self._sort = next((header['shortname'] for header in self.headers if header['character'] is not
                                       None and header['character'].upper() == ch.upper()), False)

            # reset terminal
            if has_tty:
                termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
            counter += 1
            if counter == self.args.n:
                raise SystemExit

    def get_pids_by_container(self, con_id):
        """
        Call the custom API for docker python
        :param con_id: id of the container
        :return: dict
        """

        if len(self.active_containers) == 0:
            raise ValueError("No containers running")
        f_procs = []
        # Get the name of the container by ID
        con_name = str(next((l for l in self.active_containers if l['Id'].startswith(con_id)), None)['Names'][0])
        if con_name.startswith("/"):
            con_name = con_name.replace("/", "")
        # Assemble the ps args
        ps_args = [header['ps_opt']for header in sorted(self.headers, key=itemgetter('index')) if header['ps_opt']
                   is not None and header['active']]
        con_procs = self.d.top(con_id, ps_args="-eo {}".format(",".join(ps_args)))
        # Set the column header titles one-time
        if self.titles is None:
            self.titles = con_procs['Titles']

        if con_procs['Processes'] != None:
            # Massage the information into a dict
            for proc in con_procs['Processes']:
                t_dict = {'CID': con_id,
                          'NAME': con_name}
                for place in range(0, len(proc)):
                    t_dict[self.titles[place]] = proc[place]
                f_procs.append(t_dict)
        return f_procs

    def output_top(self, sorted_info):
        """
        Primary function for output to stdout
        :param sorted_info: list of dicts containing process information
        :return: None
        """
        cols = [col for col in sorted(self.headers, key=itemgetter('index')) if col['active']]
        active_column_names = [x['shortname'] for x in cols]
        out_format = " ".join([x['field'] for x in cols])
        col_headers = []

        # Add a '*' to the column header name currently being
        # sorted on
        for col in cols:
            title = col['column']
            if col['shortname'] == self._sort:
                title += "*"
            col_headers.append(title)
        formatted_col_headers = "\033[7m" + out_format.format(*col_headers) + "\033[0m"
        # output ATOMIC TOP title
        almost_center = len(formatted_col_headers)
        center_col = '{:^%WIDTH%}'.replace("%WIDTH%", str(almost_center))
        util.write_out(center_col.format("ATOMIC TOP\n"))
        # Output the headers
        util.write_out(formatted_col_headers)
        for ps in sorted_info:
            line_out = []
            for val in active_column_names:
                line_out.append(ps[val])
            # Output the ps information
            util.write_out(out_format.format(*line_out))

    def reformat_ps_info(self, proc_info):
        """
        Takes a structure of process information and re-organizes it into
        a list of dictionaries, where each dict is a process.
        :param proc_info:
        :return: a list of dicts
        """
        # Determine if the sort field needs to reverse the order of the sort
        _reverse = next((header['sort_order'] for header in self.headers if header['shortname'] == self._sort), False)
        if self.debug:
            util.write_out("sorting on {0} and reverse is {1}".format(self._sort, _reverse))
        return sorted(proc_info, key=itemgetter(self._sort), reverse=_reverse)







#!/usr/bin/python -Es

import dbus
import time
import threading
import dbus.service
import dbus.mainloop.glib
import json
from gi.repository import GObject
import slip.dbus.service
import Atomic
from Atomic.containers import Containers
from Atomic.delete import Delete
from Atomic.diff import Diff
from Atomic.help import AtomicHelp
from Atomic.info import Info
from Atomic.install import Install
from Atomic.images import Images
from Atomic.mount import Mount
from Atomic.pull import Pull
from Atomic.run import Run
from Atomic.scan import Scan
from Atomic.stop import Stop
from Atomic.sign import Sign
from Atomic.storage import Storage
from Atomic.top import Top
from Atomic.trust import Trust
from Atomic.uninstall import Uninstall
from Atomic.verify import Verify

class atomic_dbus(slip.dbus.service.Object):
    default_polkit_auth_required = "org.atomic.readwrite"

    class Args():
        def __init__(self):
            self.all = False
            self.args = []
            self.assumeyes = True
            self.command = []
            self.compares = []
            self.container = False
            self.containers = []
            self.debug = False
            self.default_policy = None
            self.delete_targets = []
            self.devices = None
            self.diff = False
            self.display = False
            self.downgrade=False
            self.driver = None
            self.export_location = None
            self.force = False
            self.gnupghome = None
            self.graph = None
            self.heading = False
            self.hotfix = False
            self.image = None
            self.images = []
            self.import_location = None
            self.json = True
            self.keytype = None
            self.list = False
            self.live = False
            self.mountpoint = None
            self.name = None
            self.names_only = False
            self.no_files = False
            self.optional = None
            self.options = None
            self.os = None
            self.pretty = False
            self.preview = False
            self.prune = False
            self.pubkeys = []
            self.quiet = True
            self.raw = False
            self.rebase=False
            self.reboot=False
            self.reboot=False
            self.recurse = False
            self.refspec = None
            self.reg_type = None
            self.registry = None
            self.remote = False
            self.revision = None
            self.rootfs = []
            self.rpms = False
            self.save = False
            self.scan_targets = []
            self.scan_type = None
            self.scanner = None
            self.setvalues = []
            self.shared = False
            self.sign_by = None
            self.signature_path = None
            self.sigstore = None
            self.sigstoretype = None
            self.spc = False
            self.storage = None
            self.system = False
            self.truncate = False
            self.trust_type = None
            self.user = None
            self.verbose = False

    def __init__(self, *p, **k):
        super(atomic_dbus, self).__init__(*p, **k)
        self.atomic = Atomic.Atomic()
        self.tasks = []
        self.tasks_lock = threading.Lock()
        self.last_token = 0
        self.scheduler_thread = threading.Thread(target = self.Scheduler)
        self.scheduler_thread.daemon = True
        self.scheduler_thread.start()
        self.results = dict()
        self.results_lock = threading.Lock()

    def Scheduler(self):
        while True:
            current_task = None
            with self.tasks_lock:
                if(len(self.tasks) > 0):
                    current_task = self.tasks.pop(0)
            if current_task is not None:
                result = current_task[1].scan()
                with self.results_lock:
                    self.results[current_task[0]] = result
            time.sleep(1)

    def AllocateToken(self):
        with self.tasks_lock:
            self.last_token += 1
            return self.last_token

    # atomic diff section
    # The Diff method shows differences between two container images, file
    # diff or RPMS.
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='ssbbb',out_signature='s')
    def Diff(self, src, dest, rpms, no_files, names_only):
        diff = Diff()
        args = self.Args()
        args.compares = [src, dest]
        args.verbose = True
        args.no_files = no_files
        args.names_only = names_only
        args.rpms = rpms
        diff.set_args(args)
        return json.dumps(diff.diff())

    # atomic containers section
    # The ContainersList method will list all containers on the system.
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='', out_signature='s')
    def ContainersList(self):
        c = Containers()
        args = self.Args()
        args.all=True
        c.set_args(args)
        return json.dumps(c.ps())

    # atomic containers section
    # The ContainersDelete method will delete one or more containers on the system.
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='asbb', out_signature='')
    def ContainersDelete(self, containers, all_containers, force):
        c = Containers()
        args = self.Args()
        args.container = containers[0]
        args.containers = containers[1:]
        args.force = force
        args.all = all_containers
        c.set_args(args)
        return c.delete()

    # The ContainersTrim method will Discard unused blocks (fstrim) on rootfs of running containers.
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='', out_signature='')
    def ContainersTrim(self):
        c = Containers()
        return c.fstrim()

    # atomic images section
    # The ImagesHelp - Display help associated with the image
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='s', out_signature='s')
    def ImagesHelp(self, image):
        h = AtomicHelp()
        args = self.Args()
        args.image=image
        h.set_args(args)
        return h.help()

    # atomic images section
    # The ImagesInfo - display label information about an image
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='sb', out_signature='s')
    def ImagesInfo(self, image, remote):
        i = Info()
        args = self.Args()
        args.image=image
        args.remote=remote
        i.set_args(args)
        return i.info()

    # atomic images section
    # The Images method will list all installed container images on the system.
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='', out_signature='s')
    def ImagesList(self):
        images = Images()
        args = self.Args()
        args.all=True
        images.set_args(args)
        i = images.images()
        return json.dumps(i)

    # atomic containers section
    # The ImagesDelete method will delete one or more images on the system.
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='asbb', out_signature='')
    def ImagesDelete(self, images, force, remote):
        i = Delete()
        args = self.Args()
        args.delete_targets = images
        args.remote = remote
        args.force = force
        i.set_args(args)
        return i.delete_image()

    # The ImagesPrune method will delete unused 'dangling' images
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='', out_signature='')
    def ImagesPrune(self):
        d = Delete()
        return d.prune_images()

    # The ImagesPull method will pull the specified image
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='sss', out_signature='')
    def ImagePull(self, image, storage, reg_type):
        p = Pull()
        args = self.Args()
        args.image = image
        args.storage = storage
        if reg_type != "":
            args.reg_type = reg_type
        p.set_args(args)
        return p.pull_image()

    # The ImagesUpdate method downloads the latest container image.
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='sb', out_signature='')
    def ImagesUpdate(self, image, force):
        args = self.Args()
        args.image = image
        args.force = force
        self.atomic.set_args(args)
        self.atomic.update()

    # The Vulnerable method will send back information that says
    # whether or not an installed container image is vulnerable
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='', out_signature='s')
    def VulnerableInfo(self):
        args = self.Args()
        self.atomic.set_args(args)
        return self.atomic.get_all_vulnerable_info()

    # atomic install section
    # The Install method will install the specified image
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='sssbsasas', out_signature='')
    def Install(self, image, name, user, system, remote, setvalues, extra_args):
        i = Install()
        args = self.Args()
        args.image = image
        args.name = name
        args.user = user
        args.system = system
        args.remote = remote
        args.setvalues = setvalues
        args.args = extra_args
        i.set_args(args)
        return i.install()

    # atomic mount section
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='sssbb', out_signature='')
    def MountImage(self, image, mountpoint, options, live, shared):
        mount = Mount()
        mount.image = image
        mount.mountpoint = mountpoint
        args = self.Args()
        args.options = options
        args.live = live
        args.shared = shared
        self.atomic.set_args(args)
        return mount.mount()

    # atomic mount section
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='s', out_signature='')
    def UnmountImage(self, mountpoint):
        mount = Mount()
        mount.mountpoint = mountpoint
        return mount.unmount()

    # atomic run section
    # The Run method will run the specified image
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='ssbs', out_signature='')
    def Run(self, image, name, spc, command):
        r = Run()
        args = self.Args()
        args.image = image
        args.name = name
        args.spc = spc
        args.command = command
        r.set_args(args)
        return r.run()

    # atomic scan section
    # The ScanList method will return a list of all scanners.
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='',
                         out_signature= 's')
    def ScanList(self):
        scan_list = Scan()
        args = self.Args()
        scan_list.set_args(args)
        return json.dumps(scan_list.get_scanners_list())

    # The ScanSetup method will create the scan object.
    def _ScanSetup(self, scan_targets, scanner, scan_type, rootfs, _all, images, containers):
        scan = Scan()
        args = self.Args()
        scan.useTTY = False
        if scan_targets:
            args.scan_targets = scan_targets
        if scanner:
            args.scanner = scanner
        if scan_type:
            args.scan_type = scan_type
        if len(scan_targets):
            args.scan_targets = scan_targets
        args.rootfs = rootfs
        args.all = _all
        args.images = images
        args.containers = containers
        scan.set_args(args)
        return scan


    # The Scan method will return a string.
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='asssasbbb',
                         out_signature= 's')
    def Scan(self, scan_targets, scanner, scan_type, rootfs, _all, images, containers):
        return self._ScanSetup(scan_targets, scanner, scan_type, rootfs, _all, images, containers).scan()

    # The ScheduleScan method will return a token.
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='asssasbbb', out_signature= 'x')
    def ScheduleScan(self, scan_targets, scanner, scan_type, rootfs, _all, images, containers):
        scan = self._ScanSetup(scan_targets, scanner, scan_type, rootfs, _all, images, containers)
        token = self.AllocateToken()
        with self.tasks_lock:
            self.tasks.append((token, scan))
        return token

    # The GetScanResults method will determine whether or not the results for
    # the token are ready.
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='x', out_signature= 's')
    def GetScanResults(self, token):
        with self.results_lock:
            if token in self.results:
                ret = self.results[token]
                del self.results[token]
                return ret
            else:
                return ""

    # atomic sign section
    # The create a signature for images which can be used later to verify them.
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='assss', out_signature='')
    def Sign(self, images, sign_by, signature_path, gnupghome):
        sign = Sign()
        args = self.Args()
        args.images = images
        args.sign_by = sign_by
        args.signature_path = signature_path
        args.gnupghome = gnupghome
        sign.set_args(args)
        return sign.sign()


    # atomic stop section
    # The Stop method will stop the specified image
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='ssas', out_signature='')
    def Stop(self, image, name, extra_args):
        i = Stop()
        args = self.Args()
        args.image = image
        args.name = name
        args.args = extra_args
        i.set_args(args)
        return i.stop()

    # atomic storage section
    # The StorageReset method deletes all containers and images from a system.
    # Resets storage to its initial configuration.
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='', out_signature='')
    def StorageReset(self):
        storage = Storage()
        # No arguments are passed for storage_reset function
        args = self.Args()
        storage.set_args(args)
        storage.reset()

    # The StorageImport method imports all containers and their associated
    # contents from a filesystem directory.
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='ss', out_signature='')
    def StorageImport(self, graph, import_location):
        storage = Storage()
        args = self.Args()
        args.graph = graph
        args.import_location = import_location
        storage.set_args(args)
        storage.Import()

    # The StorageExport method exports all containers and their associated
    # contents into a filesystem directory.
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='ssb', out_signature='')
    def StorageExport(self, graph="/var/lib/docker", export_location="/var/lib/atomic/migrate", force = False):
        storage = Storage()
        args = self.Args()
        args.graph = graph
        args.export_location = export_location
        args.force = force
        storage.set_args(args)
        storage.Export()

    # The StorageModify method modifies the default storage setup.
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='asv', out_signature='')
    def StorageModify(self, devices=None, driver = None):
        storage = Storage()
        args = self.Args()
        if devices:
            args.devices = devices
        else:
            args.devices = []
        args.driver = driver
        storage.set_args(args)
        storage.modify()

    # atomic top section
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='ass', out_signature='s')
    def Top(self, containers, optional):
        top = Top()
        args = self.Args()
        args.containers = containers
        args.optional = optional
        top.set_args(args)
        return top.json()

    # atomic trust section
    # The TrustShow displays system trust policy
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='', out_signature='s')
    def TrustShow(self):
        trust = Trust()
        args = self.Args()
        trust.set_args(args)
        return json.dumps(trust.show_json())

    # TrustAdd adds public key trust to specific registry repository
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='ssassss', out_signature='')
    def TrustAdd(self, registry, trusttype, pubkeys, keytype, sigstore, sigstoretype):
        trust = Trust()
        args = self.Args()
        args.registry = registry
        args.pubkeys = pubkeys
        args.keytype = keytype
        args.trust_type = trusttype
        args.sigstoretype = sigstoretype
        args.sigstore = sigstore
        trust.set_args(args)
        trust.add()

    # TrustAdd removes public key trust to specific registry repository
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='ss', out_signature='')
    def TrustDelete(self, registry, sigstoretype):
        trust = Trust()
        args = self.Args()
        args.sigstoretype = sigstoretype
        args.registry = registry
        trust.set_args(args)
        trust.delete()

    # TrustDefaultPolicy sets the default container image trust for the system
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='s', out_signature='')
    def TrustDefaultPolicy(self, default_policy):
        trust = Trust()
        args = self.Args()
        args.default_policy = default_policy
        trust.set_args(args)
        return trust.modify_default()

    # atomic uninstall section
    # The Uninstall method will uninstall the specified image
    @slip.dbus.polkit.require_auth("org.atomic.readwrite")
    @dbus.service.method("org.atomic", in_signature='ssbas', out_signature='')
    def Uninstall(self, image, name, force, extra_args):
        i = Uninstall()
        args = self.Args()
        args.image = image
        args.name = name
        args.force = force
        args.args = extra_args
        i.set_args(args)
        return i.uninstall()

    # atomic upload section
    # atomic verify section
    # The Verify method takes in an image name and returns whether or not the
    # image should be updated
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='as', out_signature='s')
    def Verify(self, images):
        verifications = []
        verify = Verify()
        verify.useTTY = False
        for image in images:
            args = self.Args()
            args.image = image
            verify.set_args(args)
            verifications.append({"Image": image,
                                  "Verification": verify.verify()}) #pylint: disable=no-member
        return json.dumps(verifications)

    # atomic version section
    # The Version method takes in an image name and returns its version
    # information in a list of dicts
    @slip.dbus.polkit.require_auth("org.atomic.read")
    @dbus.service.method("org.atomic", in_signature='sb',
                         out_signature='s')
    def ImageVersion(self, image, recurse=False):
        info = Info()
        args = self.Args()
        args.image = image
        args.recurse = recurse
        info.set_args(args)
        return json.dumps(info.get_version())

if __name__ == "__main__":
    mainloop = GObject.MainLoop()
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    system_bus = dbus.SystemBus()
    busname = dbus.service.BusName("org.atomic", system_bus)
    atomic_object = atomic_dbus(system_bus, "/org/atomic/object")
    slip.dbus.service.set_mainloop(mainloop)
    mainloop.run()



from . import util
from . import Atomic
import os
from operator import itemgetter
from .atomic import AtomicError
from .syscontainers import SystemContainers
from .mount import Mount
import argparse
import shutil
import itertools
import tempfile
import subprocess
from .discovery import  RegistryInspect
from .client import no_shaw
import json


def cli(subparser, hidden=False):
    # atomic verify
    if hidden:
        verifyp = subparser.add_parser("verify", argument_default=argparse.SUPPRESS)
    else:
        verifyp = subparser.add_parser(
            "verify", help=_("verify image is fully updated"),
            epilog="atomic verify checks whether there is a newer image "
            "available and scans through all layers to see if any of "
            "the sublayers have a new version available")
    verifyp.set_defaults(_class=Verify, func='verify')
    verifyp.add_argument("image", help=_("container image"))
    verifyp.add_argument("--no-validate", default=False, dest="no_validate",
                                action="store_true",
                                help=_("disable validating system images"))
    verifyp.add_argument("--storage", default="", dest="storage",
                         help=_("Specify the storage of the image. "
                                "If not specified and there are images with the same name in "
                                "different storages, you will be prompted to specify."))
    verifyp.add_argument("-v", "--verbose", default=False,
                          action="store_true",
                          help=_("Report status of each layer"))


class Verify(Atomic):
    def __init__(self):
        super(Verify, self).__init__()
        self.debug = False

    def verify(self):
        def fix_layers(layers):
            """
            Takes the input of layers (get_layers()) and adds a key
            and value for index.  Also checks if the Tag value is not
            blank but name is, puts tag into name.
            :param layers:
            :return: updated list of layers
            """
            for layer in layers:
                layer['index'] = layers.index(layer)
                if layer['RepoTags'] and layer['Name'] is "":
                    layer['Name'] = layer['RepoTags'][0]
            return layers
        self.set_debug()

        if not self.args.storage:
            if self.is_duplicate_image(self.image):
                raise ValueError("Found more than one Image with name {}; "
                                 "please specify with --storage.".format(self.image))
            else:
                if self.syscontainers.has_image(self.image):
                    return self.verify_system_image()

                # Check if the input is an image id associated with more than one
                # repotag.  If so, error out.
                if self.is_iid():
                    self.get_fq_name(self._inspect_image())
                # The input is not an image id
                else:
                    try:
                        iid = self._is_image(self.image)
                        self.image = self.get_fq_name(self._inspect_image(iid))
                    except AtomicError:
                        self._no_such_image()

        elif self.args.storage.lower() == "ostree":
            if self.syscontainers.has_image(self.image):
                return self.verify_system_image()
            else:
                self._no_such_image()

        elif self.args.storage.lower() == "docker":
            if self.is_iid():
                self.get_fq_name(self._inspect_image())
            # The input is not an image id
            else:
                try:
                    iid = self._is_image(self.image)
                    self.image = self.get_fq_name(self._inspect_image(iid))
                except AtomicError:
                    self._no_such_image()

        else:
            raise ValueError("{} is not a valid storage".format(self.args.storage))

        layers = fix_layers(self.get_layers())
        if self.debug:
            for l in layers:
                util.output_json(l)

        uniq_names = list(set(x['Name'] for x in layers if x['Name'] != ''))
        base_images = self.get_tagged_images(uniq_names, layers)

        if not self.useTTY:
            return base_images

        if self.debug:
            for b in base_images:
                util.output_json(b)

        self.print_verify(base_images, self.image, verbose=self.args.verbose)

    def get_tagged_images(self, names, layers):
        """
        Returns a dict with image names and its tag name.
        :param names:
        :param layers:
        :return: list of sorted dicts (by index)
        """
        base_images = []
        for name in names:
            _match = next((x for x in layers if x['Name'] == name and x['RepoTags'] is not ''), None)
            registry, repo, image, tag, _ = util.Decompose(self.get_fq_image_name(_match['RepoTags'][0])).all
            tag = "latest"
            ri = RegistryInspect(registry=registry, repo=repo, image=image, tag=tag, debug=self.debug)
            ri.ping()
            remote_inspect = ri.inspect()
            release = remote_inspect.get("Labels", None).get("Release", None)
            version = remote_inspect.get("Labels", None).get("Version", None)
            if release and version:
                remote_version =  "{}-{}-{}".format(name, version, release)
            else:
                # Check if the blob exists on the registry by the ID
                remote_id = no_shaw(ri.rc.manifest_json.get("config", None).get("digest", None))
                _match['Version'] = _match['Id']
                remote_version = remote_id if remote_id is not None else ""

            _match['Remote Version']  =  remote_version
            base_images.append(_match)
        return sorted(base_images, key=itemgetter('index'))

    @staticmethod
    def _mismatch(layer):
        if layer['Version'] != layer['Remote Version'] and layer['Remote Version'] != layer['Id']:
            return "Yes"
        if layer['Version'] == '' and layer['Remote Version'] == '':
            return "!"
        return "No"


    @staticmethod
    def print_verify(base_images, image, verbose=False):
        """
        Implements a verbose printout of layers.  Can be called with
        atomic verify -v or if we detect some layer does not have
        versioning information.
        :param base_images:
        :param image:
        :return: None
        """
        def check_for_updates(base_images):
            for i in base_images:
                if Verify._mismatch(i) in ['Yes', '!']:
                    return True
            return False
        has_updates = check_for_updates(base_images)
        if has_updates or verbose:

            col = "{0:30} {1:20} {2:20} {3:1}"
            util.write_out("\n{} contains the following images:\n".format(image))
            util.write_out(col.format("NAME", "LOCAL VERSION", "REMOTE VERSION", "DIFFERS"))
            for _image in base_images:
                util.write_out(col.format(_image['Name'][:30], _image['Version'][:20], _image['Remote Version'][:20], Verify._mismatch(_image)))
            util.write_out("\n")

    def verify_system_image(self):
        manifest = self.syscontainers.get_manifest(self.image)
        name = json.loads(manifest).get('Name', self.image)
        if manifest:
            layers = SystemContainers.get_layers_from_manifest(manifest)
        else:
            layers = [self.image]
        if not getattr(self.args,"no_validate", False):
            self.validate_system_image_manifests(layers)

        if not manifest:
            return
        remote = True
        try:
            remote_manifest = self.syscontainers.get_manifest(self.image, remote=True)
            remote_layers = SystemContainers.get_layers_from_manifest(remote_manifest)
        except subprocess.CalledProcessError:
            remote_layers = []
            remote = False

        if hasattr(itertools, 'izip_longest'):
            zip_longest = getattr(itertools, 'izip_longest')
        else:
            zip_longest = getattr(itertools, 'zip_longest')

        images = []
        for local, remote in zip_longest(layers, remote_layers):
            images.append({'Name': name,
                           'Version': no_shaw(local),
                           'Id': no_shaw(local),
                           'Remote Version': no_shaw(remote),
                           'remote': remote,
                           'no_version' : True,
                           'Repo Tags': self.image,
            })

        self.print_verify(images, self.image, verbose=self.args.verbose)

    def validate_system_image_manifests(self,layers):
        """
        Validate a system image's layers against the the associated validation manifests
        created from those image layers on atomic pull.
        :param layers: list of the names of the layers to validate
        :return: None
        """
        for layer in layers:
            mismatches = self.syscontainers.validate_layer(layer)
            if len(mismatches) > 0:
                util.write_out("modifications in layer %s layer:\n" % layer)
                for m in mismatches:
                    util.write_out("file '%s' changed checksum from '%s' to '%s'" % (m["name"], m["old-checksum"], m["new-checksum"]))

    def validate_image_manifest(self):
        """
        Validates a docker image by mounting the image on a rootfs and validate that
        rootfs against the manifests that were created. Note that it won't be validated
        layer by layer.
        :param:
        :return: None
        """
        iid = self._is_image(self.image)
        manifestname = os.path.join(util.ATOMIC_VAR_LIB, "gomtree-manifests/%s.mtree" % iid)
        if not os.path.exists(manifestname):
            return
        tmpdir = tempfile.mkdtemp()
        m = Mount()
        m.args = []
        m.image = self.image
        m.mountpoint = tmpdir
        m.mount()
        r = util.validate_manifest(manifestname, img_rootfs=tmpdir, keywords="type,uid,gid,mode,size,sha256digest")
        m.unmount()
        if r.return_code != 0:
            util.write_err(r.stdout)
        shutil.rmtree(tmpdir)

    @staticmethod
    def get_gomtree_manifest(layer, root=os.path.join(util.ATOMIC_VAR_LIB, "gomtree-manifests")):
        manifestpath = os.path.join(root,"%s.mtree" % layer)
        if os.path.isfile(manifestpath):
            return manifestpath
        return None




from . import Atomic
from . import util
from . import info
from . import verify
from . import help as Help
from .mount import Mount
from .delete import Delete
import os
import sys
import json
import math
import shutil
import tempfile
import time
import argparse

def convert_size(size):
    if size > 0:
        size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
        i = int(math.floor(math.log(size, 1000)))
        p = math.pow(1000, i)
        s = round(size/p, 2) # pylint: disable=round-builtin,old-division
        if s > 0:
            return '%s %s' % (s, size_name[i])
    return '0B'

def cli(subparser):
    # atomic images
    imagesp = subparser.add_parser("images",
                                   help=_("operate on images"))
    images_subparser = imagesp.add_subparsers(title='images subcommands',
                                              description="operate on images",
                                              help='additional help')

    # atomic images delete
    delete_parser = images_subparser.add_parser("delete",
                                                help=_("mark image for deletion"),
                                                epilog="Marks image. registry garbage-collection "
                                                "when invoked will recover used disk space")
    delete_parser.set_defaults(_class=Delete, func='delete_image')

    deletegroup = delete_parser.add_mutually_exclusive_group()
    deletegroup.add_argument("-f", "--force", default=False, dest="force",
                             action="store_true",
                             help=_("Force removal of local images, even if containers based on it exist.  Default is False"))

    deletegroup.add_argument("--remote", default=False, dest="remote",
                             action="store_true",
                             help=_("Delete image from remote repository"))

    deletegroup.add_argument("--storage", default="", dest="storage",
                       help=_("Specify the storage from which to delete the image from. "
                              "If not specified and there are images with the same name in "
                              "different storages, you will be propted to specify."))

    delete_parser.add_argument("delete_targets", nargs=argparse.ONE_OR_MORE,
                               help=_("container image(s)"))

    if util.gomtree_available():
        generate_parser = images_subparser.add_parser("generate",
                                                  help=_("generate image 'manifests' if missing"))
        generate_parser.set_defaults(_class=Images, func='generate_validation_manifest')


    Help.cli(images_subparser)

    info.cli(images_subparser)

    # atomic images list
    list_parser = images_subparser.add_parser("list",
                                              help=_("list container images on your system"),
                                              epilog="atomic images by default will list all installed "
                                                     "container images on your system.")
    list_parser.set_defaults(_class=Images, func='display_all_image_info')

    list_parser.add_argument("-a", "--all", dest="all", default=False,
                             action="store_true",
                             help=_("Show all images, including intermediate images"))

    list_parser.add_argument("-f", "--filter", dest="filter", metavar="FILTER",
                             action="append",
                             help=_("Filter output based on VARIABLE=VALUE format"))

    list_parser.add_argument("-n", "--noheading", dest="heading", default=True,
                             action="store_false",
                             help=_("do not print heading when listing the images"))

    list_parser.add_argument("--no-trunc", dest="truncate", default=True,
                             action="store_false",
                             help=_("Don't truncate output"))

    list_parser.add_argument("-q", "--quiet", dest="quiet", default=False,
                             action="store_true",
                             help=_("Only display image IDs"))
    list_parser.add_argument("--json", action='store_true',dest="json", default=False,
                             help=_("print in a machine parseable form"))

    prune_parser = images_subparser.add_parser("prune",
                                               help=_("delete unused 'dangling' images"),
                                               epilog="Using the prune command, "
                                                      "will free up disk space deleting unused "
                                                      "'dangling' images")
    prune_parser.set_defaults(_class=Delete, func='prune_images')

    verify.cli(images_subparser)

    info.cli_version(images_subparser)

class Images(Atomic):
    def __init__(self):
        super(Images, self).__init__()

    def display_all_image_info(self):
        def get_col_lengths(_images):
            '''
            Determine the max length of the repository and tag names
            :param _images:
            :return: a set with len of repository and tag
            If there are no images, return 1, 1
            '''
            repo_tags = [[i["repo"], i["tag"]] for i in _images]
            if repo_tags:
                return max([len(x[0]) for x in repo_tags]) + 2,\
                       max([len(x[1]) for x in repo_tags]) + 2
            else:
                return 1, 1

        _images = self.images()
        if self.args.json:
            json.dump(_images, sys.stdout)
            return

        if len(_images) >= 0:
            _max_repo, _max_tag = get_col_lengths(_images)
            if self.args.truncate:
                _max_id = 14
            else:
                _max_id = 65

            col_out = "{0:2} {1:" + str(_max_repo) + "} {2:" + str(_max_tag) + \
                      "} {3:" + str(_max_id) + "} {4:18} {5:14} {6:10}"

            if self.args.heading and not self.args.quiet:
                util.write_out(col_out.format(" ",
                                              "REPOSITORY",
                                              "TAG",
                                              "IMAGE ID",
                                              "CREATED",
                                              "VIRTUAL SIZE",
                                              "TYPE"))

            for image in _images:
                if self.args.filter:
                    image_info = {"repo" : image['repo'], "tag" : image['tag'], "id" : image['id'],
                                  "created" : image['created'], "size" : image['virtual_size'], "type" : image['type'],
                                  "dangling": "{}".format(image['is_dangling'])}
                    if not self._filter_include_image(image_info):
                        continue
                if self.args.quiet:
                    util.write_out(image['id'])

                else:
                    indicator = ""
                    if image["is_dangling"]:
                        indicator += "*"
                    elif image["used_image"]:
                        indicator += ">"
                    if image["vulnerable"]:
                        space = " " if len(indicator) < 1 else ""
                        if util.is_python2:
                            indicator = indicator + self.skull + space
                        else:
                            indicator = indicator + str(self.skull, "utf-8") + space
                    util.write_out(col_out.format(indicator, image['repo'], image['tag'], image['id'], image['created'], image['virtual_size'], image['type']))
            util.write_out("")
            return

    def images(self):
        _images = self.get_images(get_all=self.args.all)
        all_image_info = []
        if len(_images) >= 0:
            vuln_ids = self.get_vulnerable_ids()
            all_vuln_info = json.loads(self.get_all_vulnerable_info())
            used_image_ids = [x['ImageID'] for x in self.get_containers()]
            for image in _images:
                image_dict = dict()
                if not image["RepoTags"]:
                    continue
                if ':' in image["RepoTags"][0]:
                    repo, tag = image["RepoTags"][0].rsplit(":", 1)
                else:
                    repo, tag = image["RepoTags"][0], ""
                if "Created" in image:
                    created = time.strftime("%F %H:%M", time.localtime(image["Created"]))
                else:
                    created = ""
                if "VirtualSize" in image:
                    virtual_size = convert_size(image["VirtualSize"])
                else:
                    virtual_size = ""

                image_dict["is_dangling"] = self.is_dangling(repo)
                image_dict["used_image"] = image["Id"] in used_image_ids
                image_dict["vulnerable"] = image["Id"] in vuln_ids
                image_id = image["Id"][:12] if self.args.truncate else image["Id"]
                image_type = image['ImageType']
                image_dict["repo"] = repo
                image_dict["tag"] = tag
                image_dict["id"] = image_id
                image_dict["created"] = created
                image_dict["virtual_size"] = virtual_size
                image_dict["type"] = image_type
                image_dict["image_id"] = image["ImageId"]
                if image_dict["vulnerable"]:
                    image_dict["vuln_info"] = all_vuln_info[image["Id"]]
                else:
                    image_dict["vuln_info"] = dict()

                all_image_info.append(image_dict)
            return all_image_info

    def generate_validation_manifest(self):
        """
        Generates a gomtree validation manifest for a non-system image and stores it in
        ATOMIC_VAR_LIB
        :param:
        :return: None
        """
        _images = self.get_images(get_all=True)
        for image in _images:
            atomic_var_lib = util.ATOMIC_VAR_LIB
            if not image["RepoTags"]:
                continue
            iid = image["RepoTags"][0]
            if image["ImageType"] == "system":
                continue
            if iid == "<none>:<none>" or iid == "<none>":
                continue
            if os.path.exists(os.path.join(atomic_var_lib, "gomtree-manifests/%s.mtree" % iid)):
                continue
            manifestname = os.path.join(atomic_var_lib, "gomtree-manifests/%s.mtree" % iid)
            dname = os.path.dirname(manifestname)
            if not os.path.exists(dname):
                os.makedirs(dname)
            tmpdir = tempfile.mkdtemp()
            m = Mount()
            m.args = []
            m.image = iid
            m.mountpoint = tmpdir
            m.mount()
            r = util.generate_validation_manifest(img_rootfs=tmpdir, keywords="type,uid,gid,mode,size,sha256digest")
            m.unmount()
            with open(manifestname,"w",0) as f:
                f.write(r.stdout)
            shutil.rmtree(tmpdir)

    def _filter_include_image(self, image_info):
        filterables = ["repo", "tag", "id", "created", "size", "type", "dangling"]
        for i in self.args.filter:
            var, value = str(i).split("=")
            var = var.lower()
            if var == "repository":
                var = "repo"

            if var == "image":
                var = "id"

            if var not in filterables: # Default to allowing all images through for non-existing filterable
                continue

            if value not in image_info[var].lower():
                return False

        return True



import argparse
import json
import os

from . import util
from . import Atomic
from .client import AtomicDocker
import datetime
from dateutil.parser import parse as dateparse
from docker.errors import NotFound, APIError

try:
    from subprocess import DEVNULL  # pylint: disable=no-name-in-module
except ImportError:
    DEVNULL = open(os.devnull, 'wb')

def cli(subparser):
    # atomic containers
    c = subparser.add_parser("containers",
                             help=_("operate on containers"))
    containers_subparser = c.add_subparsers(title='containers subcommands',
                                            description="operate on containers",
                                            help='additional help')
    # atomic containers delete
    delete_parser = containers_subparser.add_parser("delete",
                                                help=_("delete specified containers"))
    delete_parser.set_defaults(_class=Containers, func='delete_image')
    delete_parser.add_argument("-f", "--force", action='store_true',
                               dest="force",
                               default=False,
                               help=_("Force removal of specified running containers"))
    delete_group = delete_parser.add_mutually_exclusive_group()
    delete_group.add_argument("-a", "--all", action='store_true',dest="all",
                               default=False,
                               help=_("Delete all containers"))
    delete_group.add_argument("container", nargs='?', action="append", 
                              help=_("Specify one or more containers. Must be final arguments."))
    delete_parser.add_argument("containers", nargs=argparse.REMAINDER, 
                               help=argparse.SUPPRESS)
    delete_parser.set_defaults(_class=Containers, func='delete')


    # atomic containers list
    pss = containers_subparser.add_parser("list",
                                          help=_("list the containers"),
                                          epilog="By default this shows only the running containers.")
    pss.set_defaults(_class=Containers, func='ps_tty')
    pss.add_argument("-a", "--all", action='store_true',dest="all", default=False,
                     help=_("show all containers"))
    pss.add_argument("-f", "--filter", metavar='FILTER', action='append', dest="filter",
                     help=_("Filter output based on conditions given in the VARIABLE=VALUE form"))
    pss.add_argument("--json", action='store_true',dest="json", default=False,
                     help=_("print in a machine parseable form"))
    pss.add_argument("-n", "--noheading", dest="heading", default=True,
                     action="store_false",
                     help=_("do not print heading when listing the containers"))
    pss.add_argument("--no-trunc", action='store_false', dest="truncate", default=True,
                     help=_("Don't truncate output"))
    pss.add_argument("-q", "--quiet", action='store_true', dest="quiet", default=False,
                     help=_("Only display container IDs"))
    # atomic containers trim
    trimp = containers_subparser.add_parser("trim",
                                            help=_("discard unused blocks (fstrim) on running containers"),
                                            epilog="Discard unused blocks (fstrim) on rootfs of running containers.")
    trimp.set_defaults(_class=Containers, func='fstrim')

class Containers(Atomic):

    def fstrim(self):
        with AtomicDocker() as client:
            for container in client.containers():
                containerId = container["Id"]
                ret = self._inspect_container(name=containerId)
                pid = ret["State"]["Pid"]
                mp = "/proc/%d/root" % (pid)
                util.write_out("Trimming container id {0}".format(containerId[0:12]))
                util.check_call(["/usr/sbin/fstrim", "-v", mp], stdout=DEVNULL)
        return

    def ps_tty(self):
        all_container_info = self.ps()
        all_containers = []
        for each in all_container_info:
            if each["Type"] == "system":
                container = each["Id"]
                status = "exited"
                created = datetime.datetime.fromtimestamp(each["Created"])
                info = self.syscontainers.get_container_runtime_info(container)
                if 'status' in info:
                    status = info["status"]

                if not self.args.all and status != "running":
                    continue

                image = each['Image']
                imageId = each['ImageID']
                command = each["Command"]
                created = created.strftime("%F %H:%M") # pylint: disable=no-member
                container_info = {"type" : "system", "container" : container,
                              "image" : image, "command" : command, "image_id" : imageId,
                              "created" : created, "status" : status,
                              "runtime" : "runc", "vulnerable" : each["vulnerable"]}

                if self.args.filter:
                    if not self._filter_include_container(container_info):
                        continue
                all_containers.append(container_info)

            elif each["Type"] == "docker":
            # Collect the docker containers
                container = each["Id"]
                ret = self._inspect_container(name=container)
                status = ret["State"]["Status"]
                image = ret['Config']['Image']
                imageId = ret['Image']
                command = u' '.join(ret['Config']['Cmd']) if ret['Config']['Cmd'] else ""
                created = dateparse(ret['Created']).strftime("%F %H:%M") # pylint: disable=no-member
                container_info = {"type" : "docker", "container" : container,
                                  "image" : image, "image_id" : imageId, "command" : command,
                                  "created" : created, "status" : status,
                                  "runtime" : "Docker", "vulnerable" : each["vulnerable"]}

                if self.args.filter:
                    if not self._filter_include_container(container_info):
                        continue
                all_containers.append(container_info)

        if not all_containers:
            return

        if self.args.json:
            util.write_out(json.dumps(all_containers))
            return

        if self.args.truncate:
            max_len_container = 12
            max_len_image = 20
            max_len_command = 20
        else:
            max_len_container = max(max([len(s["container"]) for s in all_containers]), 12)
            max_len_image = max(max([len(s["image"]) for s in all_containers]), 20)
            max_len_command = max(max([len(s["command"]) for s in all_containers]), 20)

        if self.args.quiet:
            for container in all_containers:
                util.write_out(container["container"][0:max_len_container])
            return

        col_out = "{0:2} {1:%s} {2:%s} {3:%s} {4:16} {5:9} {6:10}" % (max_len_container, max_len_image, max_len_command)

        if self.args.heading:
            util.write_out(col_out.format(" ",
                                          "CONTAINER ID",
                                          "IMAGE",
                                          "COMMAND",
                                          "CREATED",
                                          "STATUS",
                                          "RUNTIME"))

        #if self.args.truncate:
        for container in all_containers:
            indicator = ""
            if container["vulnerable"]:
                if util.is_python2:
                    indicator = indicator + self.skull + " "
                else:
                    indicator = indicator + str(self.skull, "utf-8") + " "
            util.write_out(col_out.format(indicator,
                                          container["container"][0:max_len_container],
                                          container["image"][0:max_len_image],
                                          container["command"][0:max_len_command],
                                          container["created"][0:16],
                                          container["status"][0:9],
                                          container["runtime"][0:10]))
    def ps(self):
        all_containers = []
        vuln_ids = self.get_vulnerable_ids()
        all_vuln_info = json.loads(self.get_all_vulnerable_info())

        # Collect the system containers
        for i in self.syscontainers.get_containers():
            i["vulnerable"] = i['Id'] in vuln_ids
            if i["vulnerable"]:
                i["vuln_info"] = all_vuln_info[i['Id']]
            else:
                i["vuln_info"] = dict()
            all_containers.append(i)

        # Collect the docker containers
        for container in [x["Id"] for x in self.d.containers(all=self.args.all)]:
            ret = self._inspect_container(name=container)
            ret["Type"] = "docker"
            ret["vulnerable"] = ret["Image"] in vuln_ids
            if ret["vulnerable"]:
                ret["vuln_info"] = all_vuln_info[ret["Image"]]
            else:
                ret["vuln_info"] = dict()
            all_containers.append(ret)

        return all_containers

    def _filter_include_container(self, container_info):
        filterables = ["container", "image", "command", "created", "status", "runtime"]
        for j in self.args.filter:
            var, value = str(j).split("=")
            var = var.lower()

            if var == "id" or var == "containerid":
                var = "container"

            if var not in filterables: # If the filter does not exist, default to allowing all containers through
                continue

            if value not in container_info[var]:
                return False

        return True

    def delete(self):
        results = 0
        sys_targets=[]
        docker_targets=[]
        if self.args.all:
            for c in self.get_containers():
                docker_targets.append(c["Id"])
            for c in self.syscontainers.get_containers():
                sys_targets.append(c["Id"])
        else:
            if not self.args.container and len(self.args.containers) == 0:
                raise ValueError("No containers selected")
            for c in [ self.args.container ] + self.args.containers:
                if self.syscontainers.get_checkout(c):
                    sys_targets.append(c)
                else:
                    docker_targets.append(c)

        for target in docker_targets:
            try:
                self.d.remove_container(target, force=self.args.force)
            except NotFound as e:
                util.write_err("Failed to delete container {}: {}".format(target, e))
                results += 1
            except APIError as e:
                util.write_err("Failed operation for delete container {}: {}".format(target, e))
                results += 1

        for target in sys_targets:
            try:
                self.syscontainers.uninstall(target)
            except IOError as e:
                util.write_err("Failed to delete container {}: {}".format(target, e))
                results += 1
        return results
